----------------------------------------
-- Instruction Execute & Write Phase
----------------------------------------

-------------------------------------------------------------------------------
-- Notes
-- 2007.04.19
---- rs232cł̒ʐMpɏ܂B
-- 2007.01.27
---- Ə܂BSRAM̃^C~Ô͂B
-- 2007.02.11
---- USBgݍ݁BH삵ĂA܂傤B
---- ȂɂႢĂƍ̂łˁB
-- 2007.03.01
---- SRAM[hď񂾃WX^ɁAɃANZX悤ƂƁA
---- slcBƂ肠1Xe[g͂ł݂܂B
-------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_signed.all;

entity EXWCOM is
    port(CLK : in std_logic;     -- system clock
        RESET : in std_logic;    -- reset
        -- Mr.SRAM
        SRAM_RW : out std_logic;                        --sram rw
        SRAM_ADDR : out std_logic_vector(19 downto 0);  -- sram addr
        -- SRAM
        SRAM_WD : out std_logic_vector(31 downto 0); -- to sram, write data
        SRAM_RD : in std_logic_vector(31 downto 0); -- from Sram, read data
        -- IFDR
        PC : in std_logic_vector(19 downto 0);  -- instruction address(PC)
        EXPC : out std_logic_vector(19 downto 0);   -- External PC
        BUSY : out std_logic;                       -- EXW busy signal
        VALID : in std_logic;                    -- FDR valid
        OP1 : in std_logic_vector(1 downto 0);   -- opcode1
        OP2 : in std_logic_vector(3 downto 0);   -- opcode2
        REGAD : in std_logic_vector(6 downto 0);  -- desg reg addr/imm7,disp7
        IMMSEL : in std_logic;                    -- if IMM then 1 else 0
        IMM : in std_logic_vector(31 downto 0);   -- imm
        -- register file
--        REGSEL : out std_logic;                  -- select signal 1->EXW 0->FDR
        REGWE : out std_logic;                    -- regfile write enable(0->read)
        REGAD_D : out std_logic_vector(6 downto 0); -- dest reg addr
        REG_D : out std_logic_vector(31 downto 0); -- regfile write data
        REG_S : in std_logic_vector(31 downto 0); -- R2 src data
        REG_T : in std_logic_vector(31 downto 0); -- R3/R1 src data
        -- com
        SER_START : out std_logic;
        SER_WR : out std_logic;
        SER_RD : in std_logic_vector(7 downto 0);
        SER_WD : out std_logic_vector(7 downto 0);
        SER_BUSY : in std_logic;
        SER_VALID : in std_logic

);
end EXWCOM;

architecture ARC of EXWCOM is

    component ALU is
        port( OP : in std_logic_vector(3 downto 0);          --- opcode
          SRC1,SRC2 : in std_logic_vector(31 downto 0);  --- R2 -> SRC1, R3 -> SRC2
          DST : out std_logic_vector(31 downto 0));
    end component;

    component FPU is
        port( SRC1,SRC2 : in std_logic_vector(31 downto 0);
            FUNC : in std_logic_vector(3 downto 0);
            CLK : in std_logic;
            FIN : out std_logic;
            START : in std_logic;
            OUTPUT : out std_logic_vector(31 downto 0));
    end component;


    constant ALL0 : std_logic_vector(31 downto 0) := (others => '0'); -- all 0
    signal STATE : std_logic_vector(3 downto 0);   -- state
    signal REGB : std_logic_vector(6 downto 0);   -- reg read/write addr
    signal RETAD : std_logic_vector(19 downto 0);  -- return addr(->R127)
    signal EXPCB : std_logic_vector(19 downto 0);  -- exPC buffer
    signal SADB : std_logic_vector(19 downto 0); -- sram addr buffer
    signal SWD : std_logic_vector(31 downto 0);
    signal SA : std_logic_vector(19 downto 0);
    signal SMODE : std_logic;
    signal FC : std_logic;
    -- ALU
    signal ALUOP : std_logic_vector(3 downto 0);
    signal ALUS1 : std_logic_vector(31 downto 0);
    signal ALUS2 : std_logic_vector(31 downto 0);
    signal ALUD : std_logic_vector(31 downto 0);
    -- FPU
    signal FPUOP : std_logic_vector(3 downto 0);
    signal FPUS1 : std_logic_vector(31 downto 0);
    signal FPUS2 : std_logic_vector(31 downto 0);
    signal FPUD : std_logic_vector(31 downto 0);
    signal FPUS : std_logic;
    signal FPUF : std_logic;
    -- exdevice
    signal EXSTATE : std_logic_vector(3 downto 0);   -- for exdevice state machine

    signal SERWBUF1,SERWBUF2,SERWBUF3 : std_logic_vector(7 downto 0); -- ser read data buffer
    signal SEROUTB : std_logic_vector(7 downto 0); -- ser write data buffer


begin

    ALU_INST : ALU
        port map(OP => ALUOP,
                SRC1 => ALUS1,
                SRC2 => ALUS2,
                DST  => ALUD);

    FPU_INST : FPU
        port map(SRC1 => FPUS1,
                SRC2 => FPUS2,
                FUNC => FPUOP,
                CLK => CLK,
                FIN => FPUF,
                START => FPUS,
                OUTPUT => FPUD);


    EXPC <= EXPCB;
    -- sram
    SRAM_ADDR <= SA;
    SRAM_WD <= SWD;
    SRAM_RW <= SMODE;

    process(CLK,RESET)
    begin
        if RESET = '1' then
            EXPCB <= (others => '0');
            STATE <= "1110";
            REGB <= (others =>'1');
            RETAD <= (others => '0');
            -- r0 is 0 register
            REGAD_D <= (others => '0');
            REG_D <= (others => '0');
            REGWE <= '1';

        elsif CLK'event and CLK = '1' then
            case STATE is
                when "0000"=>
--                    REGSEL <= '0';
                    if VALID = '1' then
                        BUSY <= '1';
                       case OP1 is
                            when "00" => -- ALU
                                ALUOP <= OP2;
                                REGB <= REGAD;
                                ALUS1 <= REG_S;
                                if IMMSEL = '1' then
                                    ALUS2 <= IMM;
                                else
                                    ALUS2 <= REG_T;
                                end if;
                                STATE <= "0001";
                            when "01" => -- FPU
                                FC <= '1';
                                FPUS <= '1';
                                FPUOP <= OP2;
                                REGB <= REGAD;
                                FPUS1 <= REG_S;
                                FPUS2 <= REG_T;
                                STATE <= "0010";
                            when "10" => -- Memory
                                case OP2(1) is
                                    when '0' =>
                                        case OP2(3 downto 2) is
                                            when "00" => -- imm
                                                REGWE <= '1';
                                                REGAD_D <= REGAD;
                                                REG_D <= IMM(20 downto 0) & "00000000000";
                                                STATE <= "0101";
                                            when "11" => -- store
                                                if (IMMSEL xor OP2(0)) = '1' then
                                                   STATE <= "0110";
                                                   REGWE <= '0';
                                                   REGAD_D <= REGAD;
                                                   SADB <= REG_S(19 downto 0) + REG_T(19 downto 0);
                                                else
                                                   SMODE <= '0';
                                                   if IMMSEL = '1' then
                                                       SA <= REG_S(19 downto 0) + IMM(19 downto 0);
                                                   else
                                                       SA <= REG_S(19 downto 0);
                                                   end if;
                                                   SWD <= REG_T;
                                                   STATE <= "1110";
                                                end if;
                                            when "10" => -- load
                                                SMODE <= '1';
                                                case OP2(0) is
                                                    when '0'=>
                                                        case IMMSEL is
                                                            when '0' =>
                                                                SA <= REG_S(19 downto 0);    
                                                            when others =>
                                                                SA <= REG_S(19 downto 0);
                                                        end case;
                                                    when '1' =>
                                                       case IMMSEL is
                                                          when '0' =>
                                                             SA <= REG_S(19 downto 0) + REG_T(19 downto 0);
                                                          when '1' =>
                                                              SA <= REG_S(19 downto 0) + IMM(19 downto 0);
                                                          when others =>
                                                               SA <= REG_S(19 downto 0);
                                                          end case;
                                                    when others =>
                                                        SA <= REG_S(19 downto 0);
                                                   end case;
--                                                case (OP2(0) & IMMSEL) is
  --                                                  when "00" =>
    --                                                   SA <= REG_S(19 downto 0);
      --                                              when "10" =>
        --                                               SA <= REG_S(19 downto 0) + REG_T(19 downto 0);
          --                                          when "11" =>
            --                                           SA <= REG_S(19 downto 0) + IMM(19 downto 0);
              --                                      when others=>
                --                                       SA <= REG_S(19 downto 0);
                  --                              end case;
                                                STATE <= "1000";
                                                REGB <= REGAD;
                                            when others =>
                                                STATE <= "1110";
                                        end case;
                                    when '1' => -- usb or rs232c
                                         case OP2(3 downto 2) is
                                            when "00" => -- imm
                                                REGWE <= '1';
                                                REGAD_D <= REGAD;
                                                REG_D <= IMM(20 downto 0) & "00000000000";
                                                STATE <= "0101";
                                            when "11" => -- to com "out"
                                                  SEROUTB <= REG_S(7 downto 0);
                                                  STATE <= "1011";
                                                  EXSTATE <= "0000";
                                            when "10" => -- from com "in"
                                                  REGB <= REGAD;
                                                  STATE <="1011";
                                                  EXSTATE <= "1000";
                                            when others =>
                                                STATE <= "1110";
                                         end case;
                                    when others =>
                                        STATE <= "1110";
                                end case;
                            when "11" => -- Jump
                                if OP2(3 downto 2) = "00" then
                                    case OP2(1 downto 0) is
                                        when "00" => -- jump when R1 = 0
                                            if REG_T = ALL0 then
                                                --EXPCB <= PC + IMM(19 downto 0);
                                                EXPCB <= PC + IMM(19 downto 0)-1;
                                            else
                                               EXPCB <= PC;
                                                --EXPCB <= PC + 1;
                                            end if;
                                        when "01" => -- jump when R1 != 0
                                            if REG_T /= ALL0 then
                                                --EXPCB <= PC + IMM(19 downto 0);
                                                EXPCB <= PC + IMM(19 downto 0)-1;
                                            else
                                               EXPCB <= PC;
                                                --EXPCB <= PC + 1;
                                            end if;
                                        when "10" => -- jump when R1 > 0
                                            if REG_T > ALL0 then
                                                --EXPCB <= PC + IMM(19 downto 0);
                                                EXPCB <= PC + IMM(19 downto 0)-1;
                                            else
                                               EXPCB <= PC;
                                               --EXPCB <= PC + 1;
                                            end if;
                                        when "11" => -- jump when R1 >= 0
                                            if REG_T >= ALL0 then
                                                EXPCB <= PC + IMM(19 downto 0)-1;
                                                --EXPCB <= PC + IMM(19 downto 0);
                                            else
                                               EXPCB <= PC;
                                               --EXPCB <= PC + 1;
                                            end if;
                                        when others =>
                                            EXPCB <= PC;
                                            --EXPCB <= PC + 1;
                                    end case;
                                    STATE <= "0011";
                                else
                                    case OP2 is
                                        when "1100" =>
                                            --EXPCB <= PC + (REGAD(0) & IMM(18 downto 0));
                                            EXPCB <= PC + (REGAD(0) & IMM(18 downto 0))-1;
                                            STATE <= "0011";
                                        when "1101" => -- jumpl
                                            --EXPCB <= PC + (REGAD(0) & IMM(18 downto 0));
                                            EXPCB <= PC + (REGAD(0) & IMM(18 downto 0))-1;
                                            REGB <= "1111111";
                                            --RETAD <= PC;
                                            RETAD <= PC + 1;
                                            STATE <= "0100";
                                        when "0100" =>
                                            EXPCB <= REG_S(19 downto 0);
                                            --EXPCB <= REG_S(19 downto 0);
                                            STATE <= "0011";
                                        when "0101" => -- jumprl
                                            --EXPCB <= REG_S(19 downto 0);
                                            EXPCB <= REG_S(19 downto 0);
                                            REGB <= "1111111";
                                            --RETAD <= PC;
                                            RETAD <= PC + 1;
                                            STATE <= "0100";
                                        when "0110" => -- ret
                                            EXPCB <= REG_S(19 downto 0);
                                            --EXPCB <= REG_S(19 downto 0);
                                            STATE <= "0011";
                                        when "0111" =>
                                            STATE <= "1111";
                                        when others =>
                                            EXPCB <= PC;
                                            --EXPCB <= PC + 1;
                                            STATE <= "0011";
                                    end case;
                                end if;
                            when others =>
                                STATE <= "1110";
                        end case;
                    else
                        STATE <= "0000";
                    end if;
			 when "1100" => -- ALU 1.5
			 	STATE <= "0001";
                    ALUS1 <= ALUS1;
                    ALUS2 <= ALUS2;
                when "0001"=> -- ALU 2
                    REGWE <= '1';
                    REGAD_D <= REGB;
                    REG_D <= ALUD;
                    STATE <= "1110";
                when "0010" => -- FPU 2
                    if FPUF = '1' then
                        REGWE <= '1';
                        REGAD_D <= REGB;
                        REG_D <= FPUD;
                        STATE <= "1110";
                        FC <= '0';
                    elsif FC = '1' then
                        STATE <= "0010";
                        FC <= '0';
                    else
                        FPUS <= '0';
                        STATE <= "0010";
                    end if;
                    
                when "0011"=> -- Jump 2-1
                    STATE <= "1110";
                when "0100"=> -- Jump 2-2
                    REGWE <= '1';
                    REGAD_D <= REGB;
                    REG_D <= "000000000000" & RETAD;
                    STATE <= "1110";
                when "0101" =>  -- Memory IMM 2
                    STATE <= "1110";
                when "0110" =>  -- Memory Store 2
			 	STATE <= STATE + 1;
			 when "0111" =>
                    SMODE <= '0';
                    SA <= SADB;
                    SWD <= REG_T;
                    STATE <= "1110";
                when "1000" =>  -- Memory Load 2
                    STATE <= STATE + 1;
                when "1001" =>  -- Memory Load 3
                    STATE <= STATE + 1;
                when "1010" => -- Memory Load 4
                    REGWE <= '1';
                    REGAD_D <= REGB;
                    REG_D <= SRAM_RD;
                    STATE <= "1110";
                when "1011" => -- exdevice
                    REGB <= REGB;
                    case EXSTATE is
                        when "0000" =>  -- out
                            if SER_BUSY = '0' then
                                SER_START <= '1';
                                SER_WD <= SEROUTB;
                                SER_WR <= '1';
                                EXSTATE <= "0111";
                            else
                                EXSTATE <= "0000";
                            end if;
                            STATE <= "1011";
                        when "0111" =>      -- com write end
                            SER_START <= '0';
                            STATE <= "1110";
                            EXSTATE <= "0000";
                        when "1000" =>  -- in
                            if SER_BUSY = '0' then  --- com read 1-1
                                SER_START <= '1';
                                SER_WR <= '0';
                                EXSTATE <= EXSTATE + 1;
                            else
                                EXSTATE <= "1000";
                            end if;
                            STATE <= "1011";
                        when "1001" =>
                            if SER_VALID = '1' then --- com read 1-2
                                SERWBUF1 <= SER_RD;
                                EXSTATE <= EXSTATE + 1;
                            else
                                EXSTATE <= "1001";
                            end if;
                            SER_START <= '0';
                            STATE <= "1011";
                        when "1010" =>
                            if SER_BUSY = '0' then  --- com read 2-1
                                SER_START <= '1';
                                SER_WR <= '0';
                                EXSTATE <= EXSTATE + 1;
                            else
                                EXSTATE <= "1010";
                            end if;
                            STATE <= "1011";
                        when "1011" =>
                            if SER_VALID = '1' then --- com read 2-2
                                SERWBUF2 <= SER_RD;
                                EXSTATE <= EXSTATE + 1;
                            else
                                EXSTATE <= "1011";
                            end if;
                            SER_START <= '0';
                            STATE <= "1011";
                        when "1100" =>
                            if SER_BUSY = '0' then  --- com read 3-1
                                SER_START <= '1';
                                SER_WR <= '0';
                                EXSTATE <= EXSTATE + 1;
                            else
                                EXSTATE <= "1100";
                            end if;
                            STATE <= "1011";
                        when "1101" =>
                            if SER_VALID = '1' then --- com read 3-2
                                SERWBUF3 <= SER_RD;
                                EXSTATE <= EXSTATE + 1;
                            else
                                EXSTATE <= "1101";
                            end if;
                            SER_START <= '0';
                            STATE <= "1011";
                        when "1110" =>
                            if SER_BUSY = '0' then  --- com read 4-1
                                SER_START <= '1';
                                SER_WR <= '0';
                                EXSTATE <= EXSTATE + 1;
                            else
                                EXSTATE <= "1110";
                            end if;
                            STATE <= "1011";
                        when "1111" =>
                            if SER_VALID = '1' then --- com read 4-2
                                REGWE <= '1';
                                REGAD_D <= REGB;
                                REG_D <= SERWBUF1 & SERWBUF2 & SERWBUF3 & SER_RD;
                                EXSTATE <= "1000";
                                STATE <= "1110";
                            else
                                EXSTATE <= "1111";
                                STATE <= "1011";
                            end if;
                            SER_START <= '0';
                        when others =>
                            STATE <= "1110";
                            EXSTATE <= "0000";
                            SER_START <= '0';
                    end case;
                --when "1100" =>
                --    STATE <= "1110";
                when "1110" => -- next return
                    SMODE <='1';
                    REGWE <= '0';
                    BUSY <= '0';
                    FPUS <= '0';
                    STATE <= "0000";
                when "1111" => -- HALT
                    STATE <= "1111";
                when others =>
                    STATE <= "0000";
            end case;
        end if;
    end process;

end ARC;

